require 'date'

require 'lusi_api/core/api'
require 'lusi_api/core/lookup'

require 'alma_course_loader/cli/course_loader'

require 'lusi_alma_course_loader/reader'

module LUSIAlmaCourseLoader
  module CLI
    # Field extractors for LUSI course objects
    module Extractors
      # Convert
      def self.date_to_s(date)
        date.strftime('%Y-%m-%dT%H:%M:%S')
      end

      # Named field value extractors
      EXTRACTORS = {
        accessdate: proc do |_year, course, _cohort|
          date_to_s(course.student_access_date)
        end,
        accessdateutc: proc do |_year, course, _cohort|
          date_to_s(course.student_access_date_utc)
        end,
        conv: proc do |_year, course, _cohort|
          course.is_conversion_space.to_s
        end,
        copy: proc do |_year, course, _cohort|
          course.copy_course_materials_from_prev_session.to_s
        end,
        coursetype: proc { |_year, course, _cohort| course.course_type },
        createspacedate: proc do |_year, course, _cohort|
          date_to_s(course.create_space_date)
        end,
        createspacedateutc: proc do |_year, course, _cohort|
          date_to_s(course.create_space_date_utc)
        end,
        enddate: proc do |_year, _course, cohort|
          date_to_s(cohort.end_date)
        end,
        enddateutc: proc do |_year, _course, cohort|
          date_to_s(cohort.end_date_utc)
        end,
        hide: proc do |_year, course, _cohort|
          course.hide_content_on_rollover.to_s
        end,
        longtitle: proc { |_year, course, _cohort| course.display_long_title },
        shorttitle: proc do |_year, course, _cohort|
          course.display_short_title
        end,
        spacetype: proc { |_year, course, _cohort| course.space_type },
        startdate: proc do |_year, _course, cohort|
          date_to_s(cohort.start_date)
        end,
        startdateutc: proc do |_year, _course, cohort|
          date_to_s(cohort.start_date_utc)
        end,
        status: proc { |_year, course, _cohort| course.status },
        vlespaceid: proc { |_year, course, _cohort| course.lusi_vle_space_id },
        year: proc { |_year, course, _cohort| course.lusi_year_id }
      }.freeze

      EXTRACTOR_DETAILS = {
        accessdate: 'start date of student access (local time)',
        accessdateutc: 'start date of student access (UTC)',
        conv: '"true" for conversion spaces, else "false"',
        copy: '"true" if copied from previous space, else "false"',
        coursetype: 'CMOD=course module, SOS=scheme of study',
        createdate: 'creation date of VLE space (local time)',
        createdateutc: 'creation date of VLE space (UTC)',
        enddate: 'end of course date (local time)',
        enddateutc: 'end of course date (UTC)',
        hide: '"true" if content hidden on rollover, else "false"',
        longtitle: 'VLE space long title',
        shorttitle: 'VLE space short title',
        spacetype: 'SHARED=shared VLE space, SINGLE=single VLE space',
        startdate: 'start of course date (local time)',
        startdateutc: 'start of course date (UTC)',
        vlespaceid: 'LUSI VLE space identity',
        year: 'LUSI year identity'
      }.freeze
      EXTRACTOR_DETAILS.each_value(&:freeze)
    end

    # Implements the course loader command-line interface
    class CourseLoader < ::AlmaCourseLoader::CLI::CourseLoader
      include Extractors

      option %w[-p --password], 'LUSI_PASSWORD', 'the LUSI web service password'
      option %w[-u --user], 'LUSI_USER', 'the LUSI web service username'

      # Returns a hash of named field value extractor details
      # @return [Hash<String|Symbol, String>] the field extractor details
      def extractor_details
        EXTRACTOR_DETAILS
      end

      # Returns a hash of named field value extractors
      # @return [Hash<String|Symbol, Method|Proc>] the field extractors
      def extractors
        EXTRACTORS
      end

      # Creates a Reader instance to retrieve course data
      # @return [LUSIAlmaCourseLoader::Reader] the Reader instance
      def reader
        return @reader unless @reader.nil?
        api = lusi_api
        lookup = lusi_lookup(api)
        @reader = ::LUSIAlmaCourseLoader::Reader.new(
          api, lookup, *time_period_list,
          current_year: current_time_period,
          filters: filter_list
        )
      end

      # Creates a LUSI API instance
      # @return [LUSI::API::Core::API] the LUSI API instance
      def lusi_api
        api_password = password || ENV['LUSI_API_PASSWORD'] || \
                       ENV['LUSI_WS_PASSWORD']
        api_user = user || ENV['LUSI_API_USER'] || ENV['LUSI_WS_USER']
        ::LUSI::API::Core::API.new(api_user: api_user,
                                   api_password: api_password,
                                   logger: logger)
      end

      # Creates a LUSI LookupService instance
      # @param api [LUSI::API::Core::API] the LUSI API instance
      # @return [LUSI::API::Core::Lookup::LookupService] the LUSI lookup service
      def lusi_lookup(api)
        lookup = LUSI::API::Core::Lookup::LookupService.new(api || lusi_api)
        lookup.load
        lookup
      end
    end
  end
end