require 'alma_course_loader'
require 'lusi_api/core/util'
require 'lusi_api/course'
require 'lusi_api/vle'

module LUSIAlmaCourseLoader
  # Methods for handling LUSI data
  module Helpers
    include LUSI::API::Core::Util

    # Course ID year offsets
    COURSE_ID_OFFSETS = {
      next: 1,
      previous: -1,
      rollover: -1
    }.freeze

    # LUSI instructor roles
    INSTRUCTOR_ROLES = {
      'administrative staff'.to_sym => true,
      'co-course convenor'.to_sym => true,
      'course convenor'.to_sym => true,
      'partner administrative staff'.to_sym => true,
      'partner course convenor'.to_sym => true,
      'teaching staff (not timetabled)'.to_sym => true,
      'teaching staff (timetabled)'.to_sym => true,
    }.freeze

    # Returns the major department for the course
    # @param course [LUSI::API::Course::Module] the LUSI course module
    # @return [LUSI::API::Course::CourseDepartment] the major department
    def course_department(course)
      return nil unless course && course.course_departments
      # There should only be one major department
      course.course_departments.select(&:is_major_department)[0]
    end

    # Returns the Moodle course ID for a given course/cohort
    # @param course [LUSI::API::VLE::VLESpace] the course
    # @param cohort [LUSI::API::Course::Cohort] the course cohort
    # @param offset [Integer, Symbol] a numerical year offset or symbol
    #   specifying the number of years from the current course year for which
    #   the ID is required. This allows generation of course IDs for years
    #   other than the specified course.
    #   The following symbols are shorthand for numeric offsets:
    #     :previous, :rollover - equivalent to offset: -1
    #     :next                - equivalent to offset: 1
    # @return [String] the Moodle course ID
    def course_id(course, cohort, offset = nil)
      offset = COURSE_ID_OFFSETS[offset] || offset.to_i
      year = if offset != 0
               # Get the year identity
               lusi_year_identity(course.lusi_year_id, offset: offset)
             else
               course.lusi_year_id
             end
      if course.space_type == 'SHARED'
        "lusi-#{course.lusi_vle_space_id}-#{year}"
      else
        c = course.vle_space_courses[0]
        "lusi-#{c.identity.course}-#{year}-#{c.identity.cohort}"
      end
    end

    # Returns the organisation unit for a course department
    # @param course [LUSI::API::Course::Module] the LUSI course module
    # @return [LUSI::API::Organisation::Unit] the organisation unit instance of
    #   the major department for the course
    def department(course = nil)
      dept = course_department(course)
      dept ? dept.org_unit : nil
    end

    # Returns the organisation units from LUSI
    def departments
      return @departments unless @departments.nil?
      organisation = LUSI::API::Organisation::Organisation.new
      organisation.load(@lusi_api, in_use_only: false)
      @departments = {}
      organisation.each(:department) { |o| @departments[o.talis_code] = o }
      @departments
    end

    # Returns a string row field trimmed to the specified length
    # @param value [String] the field value
    # @param length [Integer] the maximum field length
    # @param empty [Object, nil] the value to return if the field value is nil
    #   or empty
    # @return [String] the formatted field value
    def field(value, length = nil, empty = nil)
      value = empty if value.nil? || value.empty?
      length && value ? value[0...length] : value
    end

    # Returns a list of instructor enrolments for the course/cohort
    # @param year [LUSI::API::Calendar::Year] the course year
    # @param course [LUSI::API::VLE::VLESpace] the course
    # @param cohort [LUSI::API::Course::Cohort] the course cohort
    # @return [Array<LUSI::API::Course::StaffModuleEnrolment>] the course
    #   instructors
    def instructor_enrolments(year, course, cohort)
      result = []
      course.vle_space_courses.each do |c|
        enrolments = LUSI::API::Course::StaffModuleEnrolment.get_instance(
          @lusi_api, @lusi_lookup,
          active_vle_space_only: false,
          course_identity: c.identity.course,
          cohort_identity: c.identity.cohort,
          require_username: true,
          year_identity: c.identity.year
        )
        e = enrolments.select do |e|
          @instructor_roles.key?(e.enrolment_role.identity)
        end
        result.concat(e)
      end
      result
    end

    # Returns course instructor roles from the :staff_course_roles lookup table
    # @return [Hash<String, LUSI::API::Person::StaffCourseRole>] the roles
    def instructor_roles
      roles = @lusi_lookup.service(:staff_course_role)
      return [] if roles.nil?
      roles.select do |_k, v|
        INSTRUCTOR_ROLES.include?(v.description.downcase.to_sym)
      end
    end

    # Returns a comma-separated list of instructor usernames
    # @param usernames [Array<String>] the instructor usernames
    # @return [String, nil] the comma-separated list of instructor usernames
    def instructor_usernames(usernames)
      result = usernames.join(',')
      result.nil? || result.empty? ? nil : result
    end

    # Returns the ancestor organisation units of the course's major department
    # @param course [LUSI::API::Course::Module] the LUSI course module
    # @return [Array<LUSI::API::Organisation::Unit>] the organisation units in
    #   furthest-to-nearest order [institution, faculty, dept]
    def org_units(course = nil)
      dept = department(course)
      result = []
      until dept.nil?
        result.unshift(dept)
        dept = dept.parent
      end
      result
    end
  end

  # Reads course information from LUSI
  class Reader < AlmaCourseLoader::Reader
    include Helpers

    # Alma processing department code
    PROCESSING_DEPARTMENT = 'Library'.freeze

    # Alma term codes
    TERMS = {
      lusi_lent: 'LENT',
      lusi_mich: 'MICHAELMAS',
      lusi_summ: 'SUMMER'
    }.freeze

    # Initializes a new Reader instance
    # @param lusi_api [LUSI::API::Core::API] the LUSI API instance
    # @param lusi_lookup [LUSI::API::Core::Lookup::LookupService] the LUSI
    #   lookup service
    # @return [void]
    def initialize(lusi_api = nil, lusi_lookup = nil, *years, **kwargs)
      # Set the LUSI API and lookup properties for use with subsequent methods
      @lusi_api = lusi_api
      @lusi_lookup = lusi_lookup
      super(*years, **kwargs)
      @departments = departments
      @instructor_roles = instructor_roles
    end

    # Returns a list of LUSI year instances
    # @param year [LUSI::API::Calendar::Year] the base year (defaults to the
    #   current academic year)
    # @param years_previous [Integer] the number of years preceding the base
    #   year
    # @param years_next [Integer] the number of years following the base year
    def years(year = nil, years_previous: nil, years_next: nil)
      years_next = years_next.to_i
      years_previous = years_previous.to_i
      result = []
      result.push(*(year.previous(years_previous))) if years_previous > 0
      result.push(*(year.next(years_next))) if years_next > 0
      result
    end

    protected

    # Returns a list of courses for the specified year
    # @param year [Integer, String, LUSI::API::Calendar::Year] the course year
    # @return [Array<LUSI::API::VLE::VLESpace>] the courses for the year
    def courses(year)
      year = year_identity(year)
      LUSI::API::VLE::VLESpace.get_instance(@lusi_api, @lusi_lookup,
                                            active_vle_space_only: false,
                                            year_identity: year)
    end

    # Returns a list of cohorts for the specified course
    # @param year [Integer, String, LUSI::API::Calendar::Year] the course year
    # @param course [LUSI::API::Course::Module] the course
    # @return [Array<LUSI::API::Course::Cohort>] the cohorts for the course
    def course_cohorts(year, course)
      # A single VLE space is always associated with a single course cohort
      nil
    end

    # Returns the current academic year
    # @return [LUSI::API::Calendar::Year] the current academic year
    def current_academic_year
      LUSI::API::Calendar::Year.get_current_academic_year(@lusi_api,
                                                          @lusi_lookup)
    end

    # Returns a list of instructor usernames for the course/cohort
    # @param year [LUSI::API::Calendar::Year] the course year
    # @param course [LUSI::API::Course::Module] the course
    # @param cohort [LUSI::API::Course::Cohort] the course cohort
    # @return [Array<String>] the course instructors
    def instructors(year, course, cohort)
      enrolments = instructor_enrolments(year, course, cohort)
      result = enrolments.map(&:username)
      result.uniq!
      result
    end

    # Populates the CSV row (array) for a specific course/cohort
    # @param data [Array<String>] the CSV row (array)
    # @param year [LUSI::API::Calendar::Year] the course year
    # @param course [LUSI::API::Course::Module] the course
    # @param cohort [LUSI::API::Course::Cohort] the course cohort
    # @param instructors [Array<LUSI::API::Course::StaffModuleEnrolment>] the
    #   course instructors
    def row_data(data, year, course, cohort, instructors)
      # The code of the major (administering) department for the course.
      # Department codes are the talis_code of LUSI::API::Organisation::Unit
      institution, faculty, department = org_units(course).map(&:talis_code)
      # Course title
      data[1] = course.display_long_title
      # Term 1-4 (data[5..8]) are left empty
      # Number of participants (data[11]) is left empty
      # Weekly hours (data[12]) is left empty
      # Instructor 1-10 (data[17..26]) are left empty
      # All instructors
      data[27] = instructor_usernames(instructors)
      # Operation (data[28]) is left empty
      row_data_course_id(data, course, cohort)
      row_data_dates(data, course, cohort)
      row_data_departments(data, department)
      row_data_ids(data, course, cohort, institution, faculty)
      row_data_rollover(data, course, cohort)
    end

    def row_data_course_id(data, course, cohort)
      # Course code
      data[0] = course_id(course, cohort)
      # Section ID - not currently used, set a default value for all courses
      data[2] = '1'
    end

    def row_data_dates(data, course, cohort)
      # Start date
      data[9] = course.start_date_utc ? course.start_date_utc.strftime('%Y-%m-%d') : nil
      # End date
      data[10] = course.end_date_utc ? course.end_date_utc.strftime('%Y-%m-%d') : nil
      # Year
      data[13] = lusi_year(course.lusi_year_id)
    end

    def row_data_departments(data, department)
      # Academic departments
      data[3] = department
      # Processing department
      data[4] = PROCESSING_DEPARTMENT
    end

    def row_data_ids(data, course, cohort, institution, faculty)
      # Searchable ID 1
      # - get the course from the short title {course}-{cohort}-{year}
      title = course.display_short_title
      data[14] = if title && title.count('-') > 1
                   title.split('-')[0...-2].join('-')
                 else
                   title
                 end
      # Searchable ID 2
      data[15] = faculty
      # Searchable ID 3
      data[16] = institution
    end

    def row_data_rollover(data, course, cohort)
      # Rollover course code
      data[29] = course_id(course, cohort, :rollover)
      # Rollover course section - use the current section ID
      data[30] = data[2]
    end

    # Returns the LUSI year identity code for the specified year. If year is
    # not specified, returns the identity of the current academic year
    def year_identity(year = nil)
      year ? lusi_year_identity(year) : current_academic_year.identity
    end
  end
end